class Fragment: 
    def __init__(self, idx, content, n,m,p): 
        self.idx=idx
        self.content=content # the encoded content of the fragment
        self.numFragments=n # number of fragments from the original message
        self.numToAssemble=m # number of fragments required for reconstruction
        self.prime=p
        
    def getIndex(self): 
        return self.idx
    
    def getContent(self): 
        return self.content
    
    def getNumFragments(self): 
        return self.numFragments
    
    def getNumAssemble(self): 
        return self.numToAssemble
    
    def getPrime(self): 
        return self.prime
    
    def __repr__(self): 
        return "{}:{}".format(self.idx,self.content)



from IDA.tools import *
from IDA.fragment import *

        
def split(file, n, m): 
    """
    Inputs:
    file: name of the file to split
    n   : number of fragments after splitting the file
    m   : minimum number of fragments required to restore the file
    Output:
    a list of n fragments (as Fragment objects)
    """

    if m>n: 
        raise ValueError("numToAssemble must be less than numFragments")
        
    if n<=0 or m<=0: 
        raise ValueError("parameters numFragments and numToAssemble must be positive ingeters")
        
    # convert file to byte strings
    original_file=open(file, "rb").read()  
    
    # find the prime number greater than n
    # all computations are done modulo p
    p = 1021 if n<1021 else nextPrime(n)
    
    # split original_file into chunks (subfiles) of length m 
    original_subfiles=[list(original_file[i:i+m]) for i in range(0,len(original_file),m)]
    
    # for the last subfile, if the length is less than m, pad the subfile with zeros 
    # to achieve final length of m
    residue = len(original_file)%m
    if residue:
        
        last_subfile=original_subfiles[-1]
        last_subfile.extend([0]*(m-residue))
    

    building_blocks=[]
    for a in range(1,n+1): 
        row = []
        elt = 1
        for i in range(m): 
            row.append(elt)
            elt=(elt*a)%p
        building_blocks.append(row)
    
    fragments=[]
    for i in range(n): 
        fragment = []
        for k in range(len(original_subfiles)): 
            fragment.append(inner_product(building_blocks[i], original_subfiles[k],p))
        fragments.append(Fragment(i+1,fragment, n, m,p))
    
    return fragments

def assemble(fragments, filename=None): 
    '''
    Input: 
    fragments : a list of Fragments objects
    filename: a String for the name of the file to write
    Output: 
    String represents the content of the original file
    If filename is given, the content is written to the file
    '''
    
    
    n, m, p  = fragments[0].getNumFragments(), fragments[0].getNumAssemble(), fragments[0].getPrime()
    if len(fragments)<m:
        raise ValueError("Number of fragments is below the minimum number of fragments required to assemble the file.")
        
    # take the first m fragments
    fragments=fragments[:m]
    building_basis = [fragment.getIndex() for fragment in fragments]
    
    inverse_building_matrix = vandermonde_inverse(building_basis,p)
    fragments_matrix = [fragment.getContent() for fragment in fragments]
    output_matrix = matrix_product(inverse_building_matrix, fragments_matrix,p)
    
    # each column of output matrix is a chunk of the original matrix
    original_subfiles=[]
    ncol = len(output_matrix[0])
    nrow = len(output_matrix)
    for c in range(ncol): 
        col = [output_matrix[r][c] for r in range(nrow)]
        original_subfiles.append(col)
    
    # remove tailing zeros in the last chunk of the original file
    last_file = original_subfiles[-1]
    i=len(last_file)-1
    while i>=0 and last_file[i]==0: 
        i-=1
    original_subfiles[-1]=last_file[:i+1]
    
    # combine the original_subfiles
    original_file=[]
    for subfile in original_subfiles: 
        original_file.extend(subfile)
    
    
    # convert original_file to its content
    original_file_content = "".join(list(map(chr, original_file)))
    
    # write the output to file
    if filename:
        open(filename, "wb").write(bytes(original_file)) 
    
    return original_file_content



import math

def isPrime(p): 
    """
    Inputs:
    p : an integer
    Output: 
    True if p is prime, False otherwise
    """
    
    if p<2: 
        return False
    if p==2: 
        return True
    if p%2==0:
        return False
    for d in range(2,math.ceil(math.sqrt(p))+1): 
        if p%d==0: 
            return False
        
        else: 
            d+=2
    return True
    
def nextPrime(p): 
    '''
    Inputs: 
    p : an integer
    Output:
    next prime greater than p
    '''
    if p%2==0: 
        p+=1
    while not isPrime(p):
        p+=2
    return p    

def inner_product(a, b, p): 
    """
    Inputs:
    a, b :  v ectors of equal length
    p: an integer, a prime number
    Output: 
    The inner product of a and b, where computations are done modulo p  
    """
    
    result=0
    for i in range(len(a)): 
        result+=(a[i]*b[i])%p
    return result
        
def matrix_product(A,B, p): 
    '''
    Inputs: 
    A, B : matrices of dimensions that are compatible for matrix multiplication
    p: an integer
    Output: 
    a matrix resulted from multiplication of A and B, where all computations are done modulo p
    '''
    
    result = []
    for i in range(len(A)):
        row=[]
        for j in range(len(B[0])): 
            ans=0
            for k in range(len(A[0])):
                ans=(ans+(A[i][k]*B[k][j])%p)%p
            row.append(ans)
        result.append(row)
    return result
    
def transpose(M): 
    result=[[0]*len(M) for _ in range(len(M))]
    for i in range(len(M)): 
        for j in range(len(M)): 
            result[i][j]=M[j][i]
    return result
    
def modulo_inverse(x,p): 
    """
    Inputs: 
    x : an integer
    p : an integer, a prime number
    Output:
    x^-1 mod p
    """
    
    t,new_t=0,1
    r,new_r=p, x
    while new_r: 
        quotient=r//new_r
        t,new_t = new_t, t-quotient*new_t
        r,new_r = new_r, r-quotient*new_r
    
    if r>1: 
        raise ValueError("{} is not invertible mod {}".format(x,p))
    if t<0:
        t=t+p
    return t
    
def elementary_symmetric_functions(m,L,p): 
    """
    Inputs: 
    m : a integer
    L : a list of integers
    p : an integer, a prime number
    Output: 
    A list `cache` where cache[i]=E(k,L,p)
    where E(k,L,p) is the sum of all the products of k distincts elements in L, where all computations are done modulo p
    """
    
    # Recursive relation: 
    # E(k,i) = E(k, L[:i+1], p)
    # E(k,i) = E(k-1,i) + L[i]*E(k-1, i-1)
    # E(1,i) = sum (L[:i+1])
    
    cache = [[0]*(len(L)+1) for _ in range(m+1)]
    for col in range(1,len(L)+1):
        cache[1][col]=cache[1][col-1]+L[col-1]
    
    
    for row in range(2,m+1): 
        for col in range(row,len(L)+1): 
            cache[row][col]=cache[row-1][col-1]*L[col-1]+cache[row][col-1]
    
    return [cache[row][-1] for row in range(0,m+1) ]
            
def vandermonde_inverse(vandermonde_basis,p): 
    '''
    Inputs: 
    vandermonde_basis : a list of integers that formed the basis of the  vandermonde matrix
    p : an integer, a prime number
    Output: 
    the inverse matrix of the vandermonde matrix (forms by vandermonde_basis), 
    where all computations are done modulo p
    '''
    # Parts of the following code implements the algorithm presents in Man, Y.-K. (2018). On computing the inverse of Vandermonde matrix. Advances in Theoretical and Applied Mathematics, 13(1), 15-21.
    
    m = len(vandermonde_basis)
    
    # cache a dictionary for modulo inverse
    inverses={}
    
    # build a list of elementary symmetric function
    # el[i] is the sum of prod of i disticnt elements in vandemonde_basis
    el = elementary_symmetric_functions(m, vandermonde_basis, p)
    
    # denominators of each row
    denominators=[]
    for i in range(m): 
        prod=1
        elt = vandermonde_basis[i]
        for j in range(m): 
            if j!=i: 
                prod=(prod*(elt-vandermonde_basis[j]))%p
        denominators.append(prod)
   
    # numerators for each row
    # perform synthetic division 
    numerators=[]
    for i in range(m): 
        row=[1]
        elt = vandermonde_basis[i]
        sign=-1
        for j in range(1,m): 
            row.append(((row[-1]*elt)%p+sign*el[j])%p)
            sign*=-1
        row.reverse()
        numerators.append(row)
    
    result=[]
    for i in range(m): 
        denominator=denominators[i]
        if denominator in inverses: 
            inv=inverses[denominator]
        else: 
            inv=modulo_inverse(denominator,p)
            inverses[denominator]=inv
        
        row=list(map(lambda x: (x*inv)%p, numerators[i]))
        result.append(row)
    return transpose(result)



import unittest
import IDA
import random

class IDATestCase(unittest.TestCase):
        
    def test_file_not_exist(self): 
        with self.assertRaises(FileNotFoundError):
            IDA.split("notafile.xyz", 10,4)
            
    def test_numFragments_error(self): 
        with self.assertRaises(ValueError): 
            IDA.split("testfile_short.txt", 0,0)
        
        with self.assertRaises(ValueError): 
            IDA.split("testfile_short.txt", 0, -1)
        
    def test_numToAssemble_error(self): 
        with self.assertRaises(ValueError): 
            IDA.split("testfile_short.txt", 10,15)  
        
    def test_short_file(self): 
        original_file=open("testfile_short.txt", "r").read()  
        fragments=IDA.split("testfile_short.txt", 10,5)
        test=True
        for count in range(100): 
            fragments_to_assemble = [fragments[i] for i in random.sample(range(10), 5)]
            output = IDA.assemble(fragments_to_assemble)
            if not output==original_file: 
                test=False
        self.assertEqual(test,True)
        
    def test_long_file(self):     
        original_file=open("testfile_long.txt", "r").read()  
        fragments=IDA.split("testfile_long.txt", 100,75)
        test=True
        for count in range(100): 
            fragments_to_assemble = [fragments[i] for i in random.sample(range(100), 75)]
            output = IDA.assemble(fragments_to_assemble)
            if not output==original_file: 
                test=False
        self.assertEqual(test,True)
    
if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(IDATestCase)
    unittest.TextTestRunner(verbosity=2).run(suite)
